/** Source code example for "A Practical Introduction to Data
    Structures and Algorithm Analysis, 3rd Edition (Java)" 
    by Clifford A. Shaffer, Prentice Hall, 2009.
    Copyright 2008 by Clifford A. Shaffer
*/

// Linked stack implementation
class LStack<E> implements Stack<E> {
  private Link<E> top;          // Pointer to first element
  private int size;		// Number of elements

  //Constructors
  public LStack() { top = null; size = 0; }
  public LStack(int size) { top = null; size = 0; }

  public void clear() { top = null; } // Reinitialize stack
  
  public void push(E it) {      // Put "it" on stack
    top = new Link<E>(it, top);
    size++;
  }

  public E pop() {              // Remove "it" from stack
    assert top != null : "Stack is empty";
    E it = top.element();
    top = top.next();
    size--;
    return it;
  }

  public E topValue() {         // Return top value
    assert top != null : "Stack is empty";
    return top.element();
  }

  public int length() { return size; } // Return length

// Extra stuff not printed in the book.

  /**
   * Generate a human-readable representation of this stack's contents
   * that looks something like this: < 1 2 3 >.
   * This method uses toString() on the individual elements.
   * @return The string representation of this stack
   */
  public String toString()
  {
    StringBuffer out = new StringBuffer((length() + 1) * 4);
    out.append("< ");
    Link<E> temp = top;
    while (temp != null) {
      out.append(temp.element());
      out.append(" ");
      temp = temp.next();
    }
    out.append(">");
    return out.toString();
  }
}
